﻿
using System;
using System.IO;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using System.Drawing;

struct subtitle_rec
{
    public int start;
    public int end;
    public int x;
    public int y;
    public int color;
    public String sub_txt;
};

namespace ass2png
{

    class Program
    {
        static String HOME;
        static subtitle_rec[] Recs = new subtitle_rec[256];
        static int Rec_count = 0; 
        static int width = 1440, height = 1080;
        static int count = 0;
        static String resultpngname;
        static String relativepngname;

        static int Main(string[] args)
        {
            StreamReader stream;
            FileStream ostream;
            String line;
            String resultxmlname;
            bool is_first_Rec = true;
            int i;

#if DEBUG
            String[] argz = { "C:\\Users\\a\\Videos\\asscutter_cut.ass" };
            args = argz;
#endif
            if (args.Length == 0)
            {
                Console.WriteLine("usage: ass2png.exe *.ass");
                Console.ReadKey();
                return 1;
            }

            HOME = AppDomain.CurrentDomain.BaseDirectory;

            if (!File.Exists(args[0]))
            {
                Console.WriteLine(String.Format("{0} not found", args[0]));
                System.Threading.Thread.Sleep(3000);
                return 1;
            }

            if (!String.Equals(Path.GetExtension(args[0]), ".ass"))
            {
                Console.WriteLine(String.Format("{0} is not ass", args[0]));
                System.Threading.Thread.Sleep(3000);
                return 1;
            }

            resultpngname = String.Format("{0}\\{1}",
                Path.GetDirectoryName(args[0]),
                Path.GetFileNameWithoutExtension(args[0])+ "_png");
            relativepngname = Path.GetFileNameWithoutExtension(args[0]) + "_png";
            if (!Directory.Exists(resultpngname))
            {
                try
                {
                    Directory.CreateDirectory(resultpngname);
                }
                catch (Exception e)
                {
                    Console.WriteLine(e.Message);
                    Console.ReadKey();
                    return 1;
                }
            }
            resultxmlname = Path.ChangeExtension(args[0], "xml");

            if (File.Exists(resultxmlname))
            {
                try
                {
                    File.Delete(resultxmlname);
                }
                catch (Exception e)
                {
                    Console.WriteLine(e.Message);
                    Console.ReadKey();
                    return 1;
                }
            }

            try
            {
                ostream = File.OpenWrite(resultxmlname);
            }
            catch (Exception e)
            {
                Console.WriteLine(e.Message);
                Console.ReadKey();
                return 1;
            }

            // Open ass file
            try
            {
                stream = File.OpenText(args[0]);
            }
            catch (Exception e)
            {
                Console.WriteLine(e.Message);
                System.Threading.Thread.Sleep(3000);
                return 1;
            }

            subtitle_rec Cur_Rec = new subtitle_rec();
            
            do
            {
                line = stream.ReadLine();
                if (line == null)   //EOF
                    break;
                if (line.StartsWith("PlayResX:"))
                    width = int.Parse(line.Substring("PlayResX:".Length));
                if (line.StartsWith("PlayResY:"))
                    height = int.Parse(line.Substring("PlayResY:".Length));

                if (line.StartsWith("Dialogue:")) 
                {
                    Int64 starti64, endi64;
                    String timepart;
                    String subpart;
                    String caption_part = "";

                    int txtidx = 0;
                    for (i = 0; i < 9; i++)     // Read thrugh 9 commas
                    {
                        txtidx = line.IndexOf(',', txtidx);
                        txtidx++;
                    }

                    subpart = line.Substring(txtidx, line.Length - txtidx);
                    timepart = line.Substring("Dialogue:".Length, txtidx - "Dialogue:".Length);
                    

                    char[] separator = { ':', ',', '.' };
                    String[] strs = timepart.Split(separator);

                    int layer = int.Parse(strs[0]);
                    int sH = int.Parse(strs[1]);
                    int sM = int.Parse(strs[2]);
                    int sS = int.Parse(strs[3]);
                    int sMs = int.Parse(strs[4]) * 10;    // since ass has 100msec unit

                    int eH = int.Parse(strs[5]);
                    int eM = int.Parse(strs[6]);
                    int eS = int.Parse(strs[7]);
                    int eMs = int.Parse(strs[8]) * 10;    // since ass has 100msec unit

                    starti64 = sH * (1000 * 60 * 60);
                    starti64 += sM * (1000 * 60);
                    starti64 += sS * (1000);
                    starti64 += sMs;

                    endi64 = eH * (1000 * 60 * 60);
                    endi64 += eM * (1000 * 60);
                    endi64 += eS * (1000);
                    endi64 += eMs;

                    Cur_Rec.start = (int)starti64;
                    Cur_Rec.end = (int)endi64;

                    int color = unchecked((int)0xFFFFFFFF); // Color.White
                    Cur_Rec.color = color;
                    Cur_Rec.x = -1;
                    Cur_Rec.y = -1;

                    if (subpart.Contains("{") && subpart.Contains("}"))
                    {
                        char[] blace = { '{', '}' };
                        String[] tag = subpart.Split(blace);

                        caption_part = tag[2];
                        caption_part = caption_part.Remove(caption_part.LastIndexOf("\\N"));
                        caption_part = caption_part.Replace("\\N", "\n");

                        if (tag[1].Contains("pos"))
                        {
                            String poscmd = tag[1].Substring(tag[1].IndexOf("pos"));
                            char[] paren = { '(', ')', ',' };
                            String[] poscmds = poscmd.Split(paren);
                            int x, y;

                            x = int.Parse(poscmds[1]);
                            y = int.Parse(poscmds[2]);
                            Cur_Rec.x = x;
                            Cur_Rec.y = y;
                        }
                        if (tag[1].Contains("c&H"))
                        {
                            String colorcmd = tag[1].Substring(tag[1].IndexOf("c&H"));
                            char[] sepalator = { '&', 'H' };
                            String[] colorcmds = colorcmd.Split(sepalator);
                            UInt32 b, g, r;
                            unchecked
                            {
                                color = (int)UInt32.Parse(colorcmds[2], System.Globalization.NumberStyles.HexNumber);
                                // bgr to rgb
                                b = ((UInt32)color & 0X00FF0000) >> 16;
                                g = ((UInt32)color & 0X0000FF00);
                                r = ((UInt32)color & 0X000000FF);
                                color = (int)((r << 16) | g | b);
                                color |= (int)0xFF000000;           // set alpha to max

                                Cur_Rec.color = color;
                                if (Cur_Rec.x == -1 || Cur_Rec.y == -1)
                                {
                                    Cur_Rec.x = width / 2;
                                    Cur_Rec.y = height;
                                }
                            }
                        }
                    }
                    else
                    {
                        caption_part = subpart;
                        caption_part = caption_part.Remove(caption_part.LastIndexOf("\\N"));
                        caption_part = caption_part.Replace("\\N", "\n");
                        Cur_Rec.x = width/2;
                        Cur_Rec.y = height;
                    }
                    Cur_Rec.sub_txt = caption_part;
                    

                    if (is_first_Rec == true)
                    {
                        Recs[0] = Cur_Rec;
                        Rec_count=1;
                        writestring(String.Format("<pgssup resolution=\"{0},{1}\" defaultoffset=\"0,810\">\n", width, height), ostream);
                        is_first_Rec = false;
                    }
                    else
                    {
                        if ((Recs[Rec_count - 1].start != Cur_Rec.start) || (Recs[Rec_count - 1].end != Cur_Rec.end))
                        {
                            Flush_Records(ostream);
                            Recs[0] = Cur_Rec;
                            Rec_count = 1;
                        }
                        else
                        {
                            Recs[Rec_count] = Cur_Rec;
                            Rec_count++;
                        }
                    }   
                }
           } while (true);
            Flush_Records(ostream);
            writestring("</pgssup>\n", ostream);
            stream.Close();
            ostream.Close();
            System.Threading.Thread.Sleep(2000);
            return 0;
        }
 
        static void Flush_Records(FileStream ostream)
        {
            Int64 starti64, endi64;
            int i;

            int sH;
            int sM;
            int sS;
            int sMs;
            int eH;
            int eM;
            int eS;
            int eMs;

            starti64 = Recs[0].start;
            endi64 = Recs[0].end;

            sH = (int)((double)starti64 / (double)(1000 * 60 * 60));
            sM = (int)((double)starti64 / (double)(1000 * 60)) % 60;
            sS = (int)((double)starti64 / (double)(1000)) % 60;
            sMs = (int)starti64 % 1000;

            eH = (int)((double)endi64 / (double)(1000 * 60 * 60));
            eM = (int)((double)endi64 / (double)(1000 * 60)) % 60;
            eS = (int)((double)endi64 / (double)(1000)) % 60;
            eMs = (int)endi64 % 1000;

            writestring("\t<subtitle starttime=", ostream);
            writestring(String.Format("\"{0:0}:{1:00}:{2:00}.{3:00}\" ", sH, sM, sS, sMs / 10), ostream);
            writestring(String.Format("endtime=\"{0:0}:{1:00}:{2:00}.{3:00}\" ", eH, eM, eS, eMs / 10), ostream);

            String target = String.Format("png{0:0000}.png", count);

            try
            {
                File.Delete(resultpngname + "\\" + target);
            }
            catch (Exception e)
            {
                Console.WriteLine(e.Message);
                Console.ReadKey();
            }
            Bitmap back = new Bitmap(width, height);
            Graphics bg = Graphics.FromImage(back); //get Graphics object of back
            back.MakeTransparent();
            bg.Clear(Color.Transparent);            //clear with Transparent color

            // Create font and brush.
            Font drawFont = new Font("MS UI Gothicl", 36.0F, FontStyle.Regular);

            // Measure string size in pixels
            SizeF str_w = bg.MeasureString(Recs[0].sub_txt, drawFont);

            if (Rec_count == 1)
            {
                // Create Bitmap fitting with the string
                Bitmap str_back = new Bitmap((int)str_w.Width+8, (int)str_w.Height+8);
                Graphics str_bg = Graphics.FromImage(str_back);

                // Create Bitmap for shadow mask
                Bitmap mask_back = new Bitmap((int)str_w.Width + 8, (int)str_w.Height + 8);
                Graphics mask_bg = Graphics.FromImage(mask_back);

                str_back.MakeTransparent();
                str_bg.Clear(Color.Transparent);
                mask_bg.Clear(Color.Black);

                SolidBrush drawBrush = new SolidBrush(unchecked(Color.FromArgb(Recs[0].color)));
                //str_bg.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.None; 
                //str_bg.TextRenderingHint = System.Drawing.Text.TextRenderingHint.SingleBitPerPixel;
                str_bg.DrawString(Recs[0].sub_txt, drawFont, drawBrush, 3, 3);
                drawBrush.Dispose();

                mask_bg.DrawString(Recs[0].sub_txt, drawFont, Brushes.Blue, 0, 0);
                mask_bg.DrawString(Recs[0].sub_txt, drawFont, Brushes.Red, 8, 8);
                mask_bg.DrawString(Recs[0].sub_txt, drawFont, Brushes.Blue, 6, 6);
                drawFont.Dispose();
                make_border(str_back, mask_back, (int)str_w.Width+8, (int)str_w.Height+8);
                try
                {
                    str_back.Save((resultpngname + "\\" + target), System.Drawing.Imaging.ImageFormat.Png);
                }
                catch (Exception e)
                {
                    Console.WriteLine(e.Message);
                    Console.ReadKey();
                }

                str_back.Dispose();
                str_bg.Dispose();
                mask_back.Dispose();
                mask_bg.Dispose();
                writestring(String.Format("offset=\"{0},{1}\" ", Recs[0].x - (int)(str_w.Width / 2), Recs[0].y - (int)str_w.Height), ostream);
                writestring(String.Format("image=\"" + relativepngname + "\\{0}\" />\n", target), ostream);
            }
            else
            {
                Point top_left = new Point(width, height);
                Point bot_right = new Point(0, 0);

                for (i = 0; i < Rec_count; i++)
                {
                    // Measure string size in pixels
                    str_w = bg.MeasureString(Recs[i].sub_txt, drawFont);

                    // convert (bottom, center) to (top, left)
                    int x = Recs[i].x - (int)(str_w.Width / 2);
                    int y = Recs[i].y - (int)str_w.Height;
                    top_left.X = (x < top_left.X) ? x : top_left.X;
                    top_left.Y = (y < top_left.Y) ? y : top_left.Y;

                    //(top, reft) to (bot, right)
                    x = x + (int)str_w.Width;
                    y = y + (int)str_w.Height;
                    bot_right.X = (x > bot_right.X) ? x : bot_right.X;
                    bot_right.Y = (y > bot_right.Y) ? y : bot_right.Y;
                }
                Bitmap target_b = new Bitmap(bot_right.X - top_left.X, bot_right.Y - top_left.Y);
                Graphics target_g = Graphics.FromImage(target_b);
                for (i = 0; i < Rec_count; i++)
                {
                    // Measure string size in pixels
                    str_w = bg.MeasureString(Recs[i].sub_txt, drawFont);


                    // Create Bitmap fitting with the string
                    Bitmap str_back = new Bitmap((int)str_w.Width + 8, (int)str_w.Height + 8);
                    Graphics str_bg = Graphics.FromImage(str_back);
                    int x, y;

                    // Create Bitmap for shadow mask
                    Bitmap mask_back = new Bitmap((int)str_w.Width + 8, (int)str_w.Height + 8);
                    Graphics mask_bg = Graphics.FromImage(mask_back);

                    str_back.MakeTransparent();
                    str_bg.Clear(Color.Transparent);
                    mask_bg.Clear(Color.Black);

                    SolidBrush drawBrush = new SolidBrush(unchecked(Color.FromArgb(Recs[i].color)));
                    str_bg.DrawString(Recs[i].sub_txt, drawFont, drawBrush, 3, 3);
                    drawBrush.Dispose();

                    mask_bg.DrawString(Recs[i].sub_txt, drawFont, Brushes.Blue, 0, 0);
                    mask_bg.DrawString(Recs[i].sub_txt, drawFont, Brushes.Red, 8, 8);
                    mask_bg.DrawString(Recs[i].sub_txt, drawFont, Brushes.Blue, 6, 6);
                    make_border(str_back, mask_back, (int)str_w.Width+8, (int)str_w.Height+8);
                    
					// convert (bottom, center) to (top, left)
                    x = Recs[i].x - (int)(str_w.Width/2);
                    y = Recs[i].y - (int)str_w.Height;
                    x -= top_left.X;
                    y -= top_left.Y;

                    target_g.DrawImage(str_back, x, y, (int)str_w.Width, (int)str_w.Height);
                    str_back.Dispose();
                    str_bg.Dispose();
                    mask_back.Dispose();
                    mask_bg.Dispose();
                }
                drawFont.Dispose();

                try
                {
                    target_b.Save((resultpngname + "\\" + target), System.Drawing.Imaging.ImageFormat.Png);
                }
                catch (Exception e)
                {
                    Console.WriteLine(e.Message);
                    Console.ReadKey();
                }
                writestring(String.Format("offset=\"{0},{1}\" ", top_left.X, top_left.Y), ostream);
                writestring(String.Format("image=\"" + relativepngname + "\\{0}\" />\n", target), ostream);
                
                target_g.Dispose();
                target_b.Dispose();
            }
            
            bg.Dispose();
            back.Dispose();
            Console.Write(String.Format("{0:00}:{1:00}:{2:00}.{3:000} --> ", sH, sM, sS, sMs));
            Console.WriteLine(String.Format("{0:00}:{1:00}:{2:00}.{3:000}", eH, eM, eS, eMs));

            count++;
        }

        static void writestring(String str, FileStream ofs)
        {
            byte[] buf = Encoding.Default.GetBytes(str);    // Encoding.Default == sjis in japanese system
            ofs.Write(buf, 0, buf.Length);
        }

        static void make_border(Bitmap bitmap, Bitmap mask, int Width, int Height)
        {
            int x, y;
            Color self;
            int rgb;
            //return;
            for (y = 0; y < Height; y++)
            {
                for (x = 0; x < Width; x++)
                {
                    unchecked
                    {
                        if (bitmap.GetPixel(x, y).A == 0)
                        {
                            self = mask.GetPixel(x, y);
                            rgb = self.ToArgb();
                            if (rgb != Color.Black.ToArgb())
                            {
                                rgb = 0;
                                if (self.B == 255)
                                    rgb = (int)((UInt32)(128) << 24);
                                else if (self.R == 255)
                                    rgb = (int)((UInt32)(96) << 24);
                                if (rgb != 0)
                                    bitmap.SetPixel(x, y, Color.FromArgb(rgb));
                            }
                        }
                    }
                }
            }
        }

        


    }   //-------------------------class Program-----------------------------
}
